#define _BSD_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <poll.h>

#include "control.h"
#include "debug.h"

#define CONTROL_PORT 10000
#define CONTROL_ADDR 0x7f000001

int control_fd;

void control_init() {
	control_fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (control_fd == -1) {
		perror("Error creating datagram socket");
		exit(EXIT_FAILURE);
	}

	struct sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(CONTROL_PORT);
	addr.sin_addr.s_addr = htonl(CONTROL_ADDR);

	int rc = bind(control_fd, (struct sockaddr*) &addr, sizeof(addr));
	if (rc == -1) {
		perror("Error binding datagram socket");
		exit(EXIT_FAILURE);
	}
}

void control_close() {
	close(control_fd);
	control_fd = -1;
}

int control_poll(int timeout_millis) {
	struct pollfd fds;
	fds.fd = control_fd;
	fds.events = POLLIN;

	int rc = poll(&fds, 1, timeout_millis);
	if (rc == -1) {
		perror("Error polling control socket");
		exit(EXIT_FAILURE);
	}

	return fds.events;
}

int control_parse(char* buffer, struct message* message);

int control_read(struct message* message) {
	char buffer[4096];
	memset(buffer, 0, sizeof(buffer));

	ssize_t rc = recv(control_fd, buffer, sizeof(buffer) - 1, MSG_DONTWAIT);
	if (rc == -1) {
		if (errno == EAGAIN) {
			return 0;
		} else {
			perror("Error receiving control packet");
			exit(EXIT_FAILURE);
		}
	}

	int parse_rc = control_parse(buffer, message);
	if (parse_rc != 0) {
		printf("Parse error: %d\n", parse_rc);
		return 0;
	}
	return 1;
}

int control_read_timeout(struct message* message, int timeout_millis) {
	int rc = control_poll(timeout_millis);
	if (rc != 0) {
		rc = control_read(message);
	}
	return rc;
}

void control_read_blocking(struct message* message) {
	char buffer[4096];
	memset(buffer, 0, sizeof(buffer));

	debug_log("Wait for control message\n");
	ssize_t rc = recv(control_fd, buffer, sizeof(buffer) - 1, 0);
	if (rc == -1) {
		perror("Error receiving control packet");
		exit(EXIT_FAILURE);
	}

	debug_log("Parsing control message\n");
	int parse_rc = control_parse(buffer, message);
	if (parse_rc != 0) {
		printf("Parse error: %d\n", parse_rc);
		exit(EXIT_FAILURE);
	}
}

int control_parse_command(char* data, command* result) {
	int rc = 0;

	if (strcmp(data, "START") == 0) {
		*result = C_START;
	} else if (strcmp(data, "PAUSE") == 0) {
		*result = C_PAUSE;
	} else if (strcmp(data, "CONTINUE") == 0) {
		*result = C_CONTINUE;
	} else {
		rc = -1;
	}

	return rc;
}

int control_parse_data_brightness(char* buffer, struct message* message);
int control_parse_data_color(char* buffer, struct message* message);
int control_parse_data_timer(char* buffer, struct message* message);
int control_parse_data_lava(char* buffer, struct message* message);
int control_parse_data_lava_dot(char* buffer, struct data_lava_dot* dot, lava_command command);

int control_parse(char* buffer, struct message* message) {
	memset(message, 0, sizeof(struct message));

	char* type = strsep(&buffer, "\n");
	if (!type) {
		return -2;
	}

	int rc = 0;
	if (strcmp(type, "RESET") == 0) {
		message->type = T_RESET;
	} else if (strcmp(type, "BRIGHTNESS") == 0) {
		message->type = T_BRIGHTNESS;
		rc = control_parse_data_brightness(buffer, message);
	} else if (strcmp(type, "COLOR") == 0) {
		message->type = T_COLOR;
		rc = control_parse_data_color(buffer, message);
	} else if (strcmp(type, "TIMER") == 0) {
		message->type = T_TIMER;
		rc = control_parse_data_timer(buffer, message);
	} else if (strcmp(type, "LAVA") == 0) {
		message->type = T_LAVA;
		rc = control_parse_data_lava(buffer, message);
	} else {
		return -3;
	}

	return rc;
}

int control_parse_int(char** buffer, int* result, int rc1, int rc2) {
	char* value = strsep(buffer, "\n");
	if (!value) {
		return rc1;
	}

	errno = 0;
	*result = strtol(value, NULL, 10);
	if (errno != 0) {
		return rc2;
	}

	return 0;
}

int control_parse_data_brightness(char* buffer, struct message* message) {
	return control_parse_int(&buffer, &message->data.brightness.value, -10, -11);
}

int control_parse_data_color(char* buffer, struct message* message) {
	int rc;

	rc = control_parse_int(&buffer, &message->data.color.red, -20, -21);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &message->data.color.green, -22, -23);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &message->data.color.blue, -24, -25);
	if (rc != 0) {
		return rc;
	}

	return 0;
}

int control_parse_data_timer(char* buffer, struct message* message) {
	char* command = strsep(&buffer, "\n");
	if (!command) {
		return -30;
	}
	if (control_parse_command(command, &message->data.timer.command)) {
		return -31;
	}

	int rc = control_parse_int(&buffer, &message->data.timer.seconds, -32, -33);
	if (rc != 0) {
		return rc;
	}

	return 0;
}

int control_parse_lava_command(char* data, lava_command* result) {
	int rc = 0;

	if (strcmp(data, "START") == 0) {
		*result = L_START;
	} else if (strcmp(data, "ADD_DOT") == 0) {
		*result = L_ADD_DOT;
	} else if (strcmp(data, "UPDATE_DOT") == 0) {
		*result = L_UPDATE_DOT;
	} else if (strcmp(data, "REMOVE_DOT") == 0) {
		*result = L_REMOVE_DOT;
	} else {
		rc = -1;
	}

	return rc;
}

int control_parse_data_lava(char* buffer, struct message* message) {
	char* command = strsep(&buffer, "\n");
	if (!command) {
		return -40;
	}
	if (control_parse_lava_command(command, &message->data.lava.lava_command)) {
		return -41;
	}

	int rc = control_parse_data_lava_dot(buffer, &message->data.lava.dot, message->data.lava.lava_command);
	if (rc != 0) {
		return rc;
	}

	return 0;
}

int control_parse_data_lava_dot(char* buffer, struct data_lava_dot* dot, lava_command command) {
	if (command == L_START) {
		return 0;
	}

	int rc;

	rc = control_parse_int(&buffer, &dot->id, -50, -51);
	if (rc != 0) {
		return rc;
	}

	if (command == L_REMOVE_DOT) {
		return 0;
	}

	rc = control_parse_int(&buffer, &dot->size, -52, -53);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &dot->red, -54, -55);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &dot->green, -56, -57);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &dot->blue, -57, -58);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &dot->period, -60, -61);
	if (rc != 0) {
		return rc;
	}

	rc = control_parse_int(&buffer, &dot->segments, -62, -63);
	if (rc != 0) {
		return rc;
	}

	return 0;
}
