#include <stdlib.h>
#include <string.h>

#include <sys/time.h>

#include <wiringPi.h>

#include "control.h"
#include "mbi5026.h"
#include "brightness.h"
#include "debug.h"
#include "util.h"
#include "timer.h"

int timer_running;
int timer_abort;

void timer_set_rgb(struct frame* frame, int l, int red, int green, int blue) {
	for (int c = 0; c < COLUMNS; c++) {
		struct led* led = &frame->columns[c].leds[l];
		led->red = red;
		led->green = green;
		led->blue = blue;
	}
}

void timer_handle_message_timer(struct data_timer* timer) {
	switch (timer->command) {
	case C_PAUSE:
		debug_log("timer_handle_message_timer: pause\n");
		timer_running = 0;
		break;
	case C_CONTINUE:
		debug_log("timer_handle_message_timer: continue\n");
		timer_running = 1;
		break;
	default:
		break;
	}
}

void timer_handle_message(struct message* message) {
	switch (message->type) {
	case T_RESET:
		debug_log("timer_handle_message: reset\n");
		timer_abort = 1;
		break;
	case T_BRIGHTNESS:
		debug_log("timer_handle_message: brightness\n");
		brightness(&message->data.brightness);
		break;
	case T_TIMER:
		timer_handle_message_timer(&message->data.timer);
		break;
	default:
		debug_log("timer_handle_message: unknown\n");
		break;
	}
}

void timer_sleep(int delay_millis) {
	struct message message;
	int remaining_delay_millis = delay_millis;

	struct timeval begin;

	do {
		util_gettimeofday(&begin);
		if (control_read_timeout(&message, remaining_delay_millis)) {
			timer_handle_message(&message);
		}

		if (timer_running) {
			int elapsed_millis = util_get_elapsed_millis_since(&begin);
			remaining_delay_millis -= elapsed_millis;
		}
	} while (remaining_delay_millis > 0);
}

void timer_loop(struct frame* frame, int r, int g, int b, int delay_millis) {
	for (int i = 0; i < LEDS; i++) {
		if (timer_abort == 1) {
			return;
		}

		timer_set_rgb(frame, i, r, g, b);
		mbi_write_frame(frame);

		timer_sleep(delay_millis);
	}
}

void timer(struct data_timer* data) {
	timer_running = 1;
	timer_abort = 0;

	int seconds = data->seconds;
	int delay_millis_green = 0;
	int delay_millis_yellow = 0;
	int delay_millis_red = 0;

	if (seconds < LEDS) {
		delay_millis_red = seconds * 1000 / LEDS;
	} else {
		delay_millis_red = 1000;
		seconds -= LEDS;
		if (seconds < 2 * LEDS) {
			delay_millis_yellow = seconds * 1000 / LEDS;
		} else {
			delay_millis_yellow = 2000;
			seconds -= 2 * LEDS;

			delay_millis_green = seconds * 1000 / LEDS;
		}
	}

	struct frame frame;
	memset(&frame, 0, sizeof(frame));

	// black -> green
	debug_log("timer: starting green phase\n");
	timer_loop(&frame, 0, 1, 0, delay_millis_green);

	// green -> yellow
	debug_log("timer: starting yellow phase\n");
	timer_loop(&frame, 1, 1, 0, delay_millis_yellow);

	// yellow -> red
	debug_log("timer: starting red phase\n");
	timer_loop(&frame, 1, 0, 0, delay_millis_red);

	// red -> black
	debug_log("timer: starting final phase\n");
	for (int i = 0; i < LEDS; i++) {
		timer_set_rgb(&frame, i, 0, 0, 0);
		mbi_write_frame(&frame);
		delay(100);
	}
}
